import os, errno
try: import subprocess
except ImportError: import compat.subprocess as subprocess
import osutil

import logging

_logger = logging.getLogger("api.util.subproc")


class ExecuteException(Exception):
    def __init__(self, subprocess):
        self.subprocess = subprocess

    def result(self):
        return u" has been terminated by exception"

    def __unicode__(self):
        return u"Subprocess " + unicode(self.subprocess) \
            + self.result() \
            + u"\n== STDOUT ====================\n" + self.subprocess.stdout \
            + u"\n== STDERR ====================\n" + self.subprocess.stderr


class SignalException(ExecuteException):
    def __init__(self, subprocess, signal):
        ExecuteException.__init__(self, subprocess)
        self.signal = signal

    def result(self):
        return u" has been terminated by signal " + unicode(self.signal)


class NonzeroExitException(ExecuteException):
    def __init__(self, subprocess, exitcode):
        ExecuteException.__init__(self, subprocess)
        self.exitcode = exitcode

    def result(self):
        return u" was finished with exit code " + unicode(self.exitcode)

    def __str__(self):
        return self.result().encode("UTF-8")

# Stdout, stderr is a file objects used to redirect subprocess stdout and stderr.
# if stdout and stderr are not specified, variables self.stdout and self.stderr will hold all subprocess output
class BaseSubprocess:
    def __init__(self, prog, args = [], input='', stdout = None, stderr = None, env = os.environ, keepStderr = False):
        self.stdout = u''
        self.stderr = u''
        self.__prog = osutil.fs_compliant_encode(prog)
        self.__args = [osutil.fs_compliant_encode(item) for item in args]
        self.__env = env
        self.__input = input
        self.__stdin = None
        self.__stdout = stdout
        self.__stderr = stderr
        self.__pid = None
        self.__retcode = -1
        self.__finished = False
        self.__keepStderr = keepStderr
        self.__forkexec()

    def __get_cmd(self):
        return [self.__prog] + self.__args

    def get_cmd(self):
        cmd = self.__get_cmd()
        return osutil.args_separator.join(cmd)

    def get_pid(self):
        return self.__pid

    def kill(self, signo):
        """Kills the associated process.
        Kills the associated OS process. wait() should be used to
        collect the process death information.
        """
        if self.__finished:
            _logger.info(u"Skipped killing finished process " + unicode(self.__pid))
        else:
            _logger.info(u"Killing process " + unicode(self.__pid))
            try:
                osutil.kill(self.__pid, signo)
            except:
                pass

    def __forkexec(self):
        self.__stdin, self.__stdinwrite = os.pipe()
        self.__writeEncodeUtf8(self.__stdinwrite, self.__input)
        os.close(self.__stdinwrite)
        
        if not self.__stdout:
            self.__stdout = subprocess.PIPE
        if not self.__stderr:
            self.__stderr = subprocess.PIPE
        self.__process = subprocess.Popen(self.__get_cmd(), stdin = self.__stdin , stdout = self.__stdout, stderr = self.__stderr, env = self.__env)
        self.__pid = self.__process.pid

    def __unicode__(self):
        return u"<subprocess[%d] %s>" % ( self.__pid, repr(self.get_cmd()) )

    def __writeEncodeUtf8(self, fd, val):
        if isinstance(val, unicode):
            val = val.encode("UTF-8")
        try:
            os.write(fd, val)
        except OSError, e:
            if e.errno != errno.EPIPE:
                raise

    def __decodeUtf8(self, fd, fd_out):
        val = os.read(fd)
        try:
            val = val.decode("UTF-8")
        except UnicodeError:
            val = val.decode("Latin-1")
        try:
            os.write(fd_out, val)
        except OSError, e:
            if e.errno != errno.EPIPE:
                raise

    def __fetch_data(self):
        child_stdout, child_stderr = self.__process.communicate()
        if child_stdout:
            try:
                self.stdout = child_stdout.decode("UTF-8")
            except UnicodeError:
                self.stdout = child_stdout.decode("Latin-1")

        if child_stderr and not self.__keepStderr:
            try:
                self.stderr = child_stderr.decode("UTF-8")
            except UnicodeError:
                self.stderr = child_stderr.decode("Latin-1")

    def run(self):
        self.__fetch_data()
        self.wait()

    def runAsync(self):
        pass

    def wait(self):
        self.__retcode = self.__process.wait()
        self.__finished = True
        if self.__retcode !=  0:
            raise NonzeroExitException(self, self.__retcode)


class Subprocess(BaseSubprocess):
    def run(self):
        _logger.info(u"Executing " + unicode(self))
        BaseSubprocess.run(self)

    def runAsync(self):
        _logger.info(u"Executing asynchronously " + unicode(self))
        BaseSubprocess.runAsync(self)

    def wait(self):
        try:
            BaseSubprocess.wait(self)
        except ExecuteException, e:
            _logger.info(unicode(e))
            raise
        else:
            _logger.info(u"Execution of " + unicode(self) + u" finished successfully.")


class CmdLine(object):
    def __init__(self, filename, process_handler = None, args = [], vars = os.environ, stdin = "", stdout = None, stderr = None):
        self.filename = filename
        self.args = args
        self.vars = vars.copy()
        self.stdin = stdin
        self.stdout = stdout
        self.process_handler = process_handler
        self.stderr = stderr
        self.proc = None

    def arg(self, arg):
        self.args = self.args + [arg]
        return self

    def get_cmd(self):
        if self.proc is not None:
            return osutil.fs_compliant_decode(self.proc.get_cmd())
        else:
            return None

    def var(self, varname, varvalue):
        if isinstance(varvalue, unicode):
            self.vars[varname] = varvalue.encode("UTF-8")
        else:
            self.vars[varname] = varvalue
        return self

    def spawn(self, keepStderr = False):
        self.proc = Subprocess(self.filename,
                               self.args,
                               self.stdin,
                               stdout = self.stdout,
                               stderr = self.stderr,
                               env = self.vars,
                               keepStderr = keepStderr)
        if self.process_handler:
            self.process_handler.setProcess(self.proc)

        self.proc.run()

        if not keepStderr:
            self.stderr = self.proc.stderr

        if self.process_handler:
            self.process_handler.setProcess(None)

        return self.proc

    def asyncSpawn(self, keepStderr = False):
        self.proc = Subprocess(self.filename,
                               self.args,
                               self.stdin,
                               stdout = self.stdout,
                               stderr = self.stderr,
                               env = self.vars,
                               keepStderr = keepStderr)
        if self.process_handler:
            self.process_handler.setProcess(self.proc)

        self.proc.runAsync()
        return self.proc
